"""Clip to fit in gamut."""
from ... import util
from ... spaces import Angle, GamutBound
from ..gamut import Fit


class Clip(Fit):
    """Clip gamut mapping class."""

    @staticmethod
    def name():
        """Get name."""

        return "clip"

    @staticmethod
    def fit(color):
        """Gamut clipping."""

        channels = util.no_nan(color.coords())
        gamut = color._space.RANGE
        fit = []

        for i, value in enumerate(channels):
            a, b = gamut[i]
            is_bound = isinstance(gamut[i], GamutBound)

            # Wrap the angle. Not technically out of gamut, but we will clean it up.
            if isinstance(a, Angle) and isinstance(b, Angle):
                fit.append(value % 360.0)
                continue

            # These parameters are unbounded
            if not is_bound:  # pragma: no cover
                # Will not execute unless we have a space that defines some coordinates
                # as bound and others as not. We do not currently have such spaces.
                a = None
                b = None

            # Fit value in bounds.
            fit.append(util.clamp(value, a, b))
        return fit



"""Delta E 76."""
from ..distance import DeltaE
from . import distance_euclidean


class DE76(DeltaE):
    """Delta E 76 class."""

    @staticmethod
    def name():
        """Name of method."""

        return "76"

    @staticmethod
    def distance(color, sample, **kwargs):
        """
        Delta E 1976 color distance formula.

        http://www.brucelindbloom.com/index.html?Eqn_DeltaE_CIE76.html

        Basically this is Euclidean distance in the Lab space.
        """

        # Equation (1)
        return distance_euclidean.distance(color, sample, space="lab")



"""XYZ D65 class."""
from ..spaces import RE_DEFAULT_MATCH, GamutUnbound
from .xyz import XYZ
import re


class XYZD65(XYZ):
    """XYZ D65 class."""

    SPACE = "xyz-d65"
    SERIALIZE = ("--xyz-d65",)
    CHANNEL_NAMES = ("x", "y", "z", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    RANGE = (
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0])
    )

    @classmethod
    def _to_xyz(cls, parent, xyzd65):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, xyzd65)

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)



"""LCH class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Cylindrical, Angle, Percent
from .luv import Luv
from .. import util
import re
import math

ACHROMATIC_THRESHOLD = 0.0000000002


def luv_to_lchuv(luv):
    """Luv to Lch(uv)."""

    l, u, v = luv

    c = math.sqrt(u ** 2 + v ** 2)
    h = math.degrees(math.atan2(v, u))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c < ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


def lchuv_to_luv(lchuv):
    """Lch(uv) to Luv."""

    l, c, h = lchuv
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return (
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    )


class Lchuv(Cylindrical, Space):
    """Lch(uv) class."""

    SPACE = "lchuv"
    SERIALIZE = ("--lchuv",)
    CHANNEL_NAMES = ("l", "c", "h", "alpha")
    CHANNEL_ALIASES = {
        "lightness": "l",
        "chroma": "c",
        "hue": "h"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"

    RANGE = (
        GamutUnbound([Percent(0), Percent(100.0)]),
        GamutUnbound([0.0, 176.0]),
        GamutUnbound([Angle(0.0), Angle(360.0)]),
    )

    @property
    def l(self):
        """Lightness."""

        return self._coords[0]

    @l.setter
    def l(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def c(self):
        """Chroma."""

        return self._coords[1]

    @c.setter
    def c(self, value):
        """chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def h(self):
        """Hue."""

        return self._coords[2]

    @h.setter
    def h(self, value):
        """Shift the hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords, alpha

    @classmethod
    def _to_luv(cls, parent, lchuv):
        """To Luv."""

        return lchuv_to_luv(lchuv)

    @classmethod
    def _from_luv(cls, parent, luv):
        """To Luv."""

        return luv_to_lchuv(luv)

    @classmethod
    def _to_xyz(cls, parent, lchuv):
        """To XYZ."""

        return Luv._to_xyz(parent, cls._to_luv(parent, lchuv))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_luv(parent, Luv._from_xyz(parent, xyz))



"""
JzCzhz class.

https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Cylindrical, Angle, OptionalPercent
from .jzazbz import Jzazbz
from .. import util
import re
import math

ACHROMATIC_THRESHOLD = 0.0003


def jzazbz_to_jzczhz(jzazbz):
    """Jzazbz to JzCzhz."""

    jz, az, bz = jzazbz

    cz = math.sqrt(az ** 2 + bz ** 2)
    hz = math.degrees(math.atan2(bz, az))

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if cz < ACHROMATIC_THRESHOLD:
        hz = util.NaN

    return [jz, cz, util.constrain_hue(hz)]


def jzczhz_to_jzazbz(jzczhz):
    """JzCzhz to Jzazbz."""

    jz, cz, hz = jzczhz
    hz = util.no_nan(hz)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if cz < 0.0:
        cz = 0.0

    return (
        jz,
        cz * math.cos(math.radians(hz)),
        cz * math.sin(math.radians(hz))
    )


class JzCzhz(Cylindrical, Space):
    """
    JzCzhz class.

    https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-25-13-15131&id=368272
    """

    SPACE = "jzczhz"
    SERIALIZE = ("--jzczhz",)
    CHANNEL_NAMES = ("jz", "cz", "hz", "alpha")
    CHANNEL_ALIASES = {
        "lightness": "jz",
        "chroma": "cz",
        "hue": "hz"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    RANGE = (
        GamutUnbound([OptionalPercent(0), OptionalPercent(1)]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([Angle(0.0), Angle(360.0)]),
    )

    @property
    def jz(self):
        """Jz."""

        return self._coords[0]

    @jz.setter
    def jz(self, value):
        """Set jz."""

        self._coords[0] = self._handle_input(value)

    @property
    def cz(self):
        """Chroma."""

        return self._coords[1]

    @cz.setter
    def cz(self, value):
        """Set chroma."""

        self._coords[1] = self._handle_input(value)

    @property
    def hz(self):
        """Hue."""

        return self._coords[2]

    @hz.setter
    def hz(self, value):
        """Set hue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] < ACHROMATIC_THRESHOLD:
            coords[2] = util.NaN
        return coords, alpha

    @classmethod
    def _to_jzazbz(cls, parent, jzczhz):
        """To Jzazbz."""

        return jzczhz_to_jzazbz(jzczhz)

    @classmethod
    def _from_jzazbz(cls, parent, jzazbz):
        """From Jzazbz."""

        return jzazbz_to_jzczhz(jzazbz)

    @classmethod
    def _to_xyz(cls, parent, jzczhz):
        """To XYZ."""

        return Jzazbz._to_xyz(parent, cls._to_jzazbz(parent, jzczhz))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_jzazbz(parent, Jzazbz._from_xyz(parent, xyz))



"""
Delta E 99o.

https://de.wikipedia.org/wiki/DIN99-Farbraum
"""
from ..distance import DeltaE
from . import distance_euclidean


class DE99o(DeltaE):
    """Delta E 99o class."""

    @staticmethod
    def name():
        """Name of method."""

        return "99o"

    @staticmethod
    def distance(color, sample, **kwargs):
        """Delta E 99o color distance formula."""

        # Equation (1)
        return distance_euclidean.distance(color, sample, space="din99o")



"""Lch D65 class."""
from ..spaces import RE_DEFAULT_MATCH
from .lab_d65 import LabD65
from .lch.base import LchBase, lch_to_lab, lab_to_lch
import re


class LchD65(LchBase):
    """Lch D65 class."""

    SPACE = "lch-d65"
    SERIALIZE = ("--lch-d65",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def _to_lab_d65(cls, parent, lchd65):
        """To Lab."""

        return lch_to_lab(lchd65)

    @classmethod
    def _from_lab_d65(cls, parent, labd65):
        """To Lab."""

        return lab_to_lch(labd65)

    @classmethod
    def _to_xyz(cls, parent, lch):
        """To XYZ."""

        return LabD65._to_xyz(parent, cls._to_lab_d65(parent, lch))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_lab_d65(parent, LabD65._from_xyz(parent, xyz))



"""SRGB color class."""
from ...spaces import RE_DEFAULT_MATCH, Space, GamutBound, OptionalPercent
from ..xyz import XYZ
from ... import util
import re
import math

RGB_TO_XYZ = [
    [0.4123907992659593, 0.357584339383878, 0.18048078840183432],
    [0.21263900587151024, 0.715168678767756, 0.07219231536073373],
    [0.01933081871559182, 0.11919477979462598, 0.9505321522496608]
]

XYZ_TO_RGB = [
    [3.2409699419045226, -1.537383177570094, -0.49861076029300355],
    [-0.9692436362808796, 1.8759675015077202, 0.04155505740717562],
    [0.055630079696993635, -0.2039769588889765, 1.0569715142428784]
]


def lin_srgb_to_xyz(rgb):
    """
    Convert an array of linear-light sRGB values to CIE XYZ using sRGB's own white.

    D65 (no chromatic adaptation)
    """

    return util.dot(RGB_TO_XYZ, rgb)


def xyz_to_lin_srgb(xyz):
    """Convert XYZ to linear-light sRGB."""

    return util.dot(XYZ_TO_RGB, xyz)


def lin_srgb(rgb):
    """
    Convert an array of sRGB values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i < 0.04045:
            result.append(i / 12.92)
        else:
            result.append(math.copysign(((abs_i + 0.055) / 1.055) ** 2.4, i))
    return result


def gam_srgb(rgb):
    """
    Convert an array of linear-light sRGB values in the range 0.0-1.0 to gamma corrected form.

    https://en.wikipedia.org/wiki/SRGB
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        abs_i = abs(i)
        if abs_i > 0.0031308:
            result.append(math.copysign(1.055 * (util.nth_root(abs_i, 2.4)) - 0.055, i))
        else:
            result.append(12.92 * i)
    return result


class SRGB(Space):
    """SRGB class."""

    SPACE = "srgb"
    # In addition to the current gamut, check HSL as it is much more sensitive to small
    # gamut changes. This is mainly for a better user experience. Colors will still be
    # mapped/clipped in the current space, unless specified otherwise.
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE, channels=3))
    CHANNEL_NAMES = ("r", "g", "b", "alpha")
    CHANNEL_ALIASES = {
        "red": 'r',
        "green": 'g',
        "blue": 'b'
    }
    WHITE = "D65"

    RANGE = (
        GamutBound([OptionalPercent(0.0), OptionalPercent(1.0)]),
        GamutBound([OptionalPercent(0.0), OptionalPercent(1.0)]),
        GamutBound([OptionalPercent(0.0), OptionalPercent(1.0)])
    )

    @property
    def r(self):
        """Adjust red."""

        return self._coords[0]

    @r.setter
    def r(self, value):
        """Adjust red."""

        self._coords[0] = self._handle_input(value)

    @property
    def g(self):
        """Adjust green."""

        return self._coords[1]

    @g.setter
    def g(self, value):
        """Adjust green."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self):
        """Adjust blue."""

        return self._coords[2]

    @b.setter
    def b(self, value):
        """Adjust blue."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """SRGB to XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_srgb_to_xyz(lin_srgb(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """XYZ to SRGB."""

        return gam_srgb(xyz_to_lin_srgb(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))



"""XYZ class."""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound
import re


class XYZ(Space):
    """XYZ class."""

    SPACE = "xyz"
    SERIALIZE = ("xyz", "--xyz-d50")
    CHANNEL_NAMES = ("x", "y", "z", "alpha")
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"

    RANGE = (
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0]),
        GamutUnbound([0.0, 1.0])
    )

    @property
    def x(self):
        """X channel."""

        return self._coords[0]

    @x.setter
    def x(self, value):
        """Shift the X."""

        self._coords[0] = self._handle_input(value)

    @property
    def y(self):
        """Y channel."""

        return self._coords[1]

    @y.setter
    def y(self, value):
        """Set Y."""

        self._coords[1] = self._handle_input(value)

    @property
    def z(self):
        """Z channel."""

        return self._coords[2]

    @z.setter
    def z(self, value):
        """Set Z channel."""

        self._coords[2] = self._handle_input(value)



"""HSL class."""
from ...spaces import Space, RE_DEFAULT_MATCH, Angle, Percent, GamutBound, Cylindrical
from ..srgb.base import SRGB
from ... import util
import re


def srgb_to_hsl(rgb):
    """SRGB to HSL."""

    r, g, b = rgb
    mx = max(rgb)
    mn = min(rgb)
    h = util.NaN
    s = 0.0
    l = (mn + mx) / 2
    c = mx - mn

    if c != 0.0:
        if mx == r:
            h = (g - b) / c
        elif mx == g:
            h = (b - r) / c + 2.0
        else:
            h = (r - g) / c + 4.0
        s = 0 if l == 0 or l == 1 else (mx - l) / min(l, 1 - l)
        h *= 60.0
        if s == 0:
            h = util.NaN

    return util.constrain_hue(h), s * 100, l * 100


def hsl_to_srgb(hsl):
    """
    HSL to RGB.

    https://en.wikipedia.org/wiki/HSL_and_HSV#HSL_to_RGB_alternative
    """

    h, s, l = hsl
    h = util.no_nan(h)
    h = h % 360
    s /= 100.0
    l /= 100.0

    def f(n):
        """Calculate the channels."""
        k = (n + h / 30) % 12
        a = s * min(l, 1 - l)
        return l - a * max(-1, min(k - 3, 9 - k, 1))

    return f(0), f(8), f(4)


class HSL(Cylindrical, Space):
    """HSL class."""

    SPACE = "hsl"
    SERIALIZE = ("--hsl",)
    CHANNEL_NAMES = ("h", "s", "l", "alpha")
    CHANNEL_ALIASES = {
        "hue": "h",
        "saturation": "s",
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"
    GAMUT_CHECK = "srgb"

    RANGE = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    @property
    def h(self):
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def s(self):
        """Saturation channel."""

        return self._coords[1]

    @s.setter
    def s(self, value):
        """Saturate or unsaturate the color by the given factor."""

        self._coords[1] = self._handle_input(value)

    @property
    def l(self):
        """Lightness channel."""

        return self._coords[2]

    @l.setter
    def l(self, value):
        """Set lightness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] == 0:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def _to_srgb(cls, parent, hsl):
        """To sRGB."""

        return hsl_to_srgb(hsl)

    @classmethod
    def _from_srgb(cls, parent, rgb):
        """From sRGB."""

        return srgb_to_hsl(rgb)

    @classmethod
    def _to_xyz(cls, parent, hsl):
        """To XYZ."""

        return SRGB._to_xyz(parent, cls._to_srgb(parent, hsl))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_srgb(parent, SRGB._from_xyz(parent, xyz))



"""
Luv class.

https://en.wikipedia.org/wiki/CIELUV
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, Percent, WHITES, Labish
from .xyz import XYZ
from .. import util
import re


def xyz_to_luv(xyz, white):
    """XYZ to Luv."""

    u, v = util.xyz_to_uv(xyz)
    w_xyz = util.xy_to_xyz(WHITES[white])
    un, vn = util.xyz_to_uv(w_xyz)

    y = xyz[1] / w_xyz[1]
    l = 116 * util.nth_root(y, 3) - 16 if y > ((6 / 29) ** 3) else ((29 / 3) ** 3) * y

    return [
        l,
        13 * l * (u - un),
        13 * l * (v - vn),
    ]


def luv_to_xyz(luv, white):
    """Luv to XYZ."""

    l, u, v = luv
    xyz = util.xy_to_xyz(WHITES[white])
    un, vn = util.xyz_to_uv(xyz)

    if l != 0:
        up = (u / (13 * l)) + un
        vp = (v / (13 * l)) + vn
    else:
        up = vp = 0

    y = xyz[1] * ((l + 16) / 116) ** 3 if l > 8 else xyz[1] * l * ((3 / 29) ** 3)

    if vp != 0:
        x = y * ((9 * up) / (4 * vp))
        z = y * ((12 - 3 * up - 20 * vp) / (4 * vp))
    else:
        x = z = 0

    return [x, y, z]


class Luv(Labish, Space):
    """Oklab class."""

    SPACE = "luv"
    SERIALIZE = ("--luv",)
    CHANNEL_NAMES = ("l", "u", "v", "alpha")
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D50"

    RANGE = (
        GamutUnbound([Percent(0), Percent(100.0)]),
        GamutUnbound([-175.0, 175.0]),
        GamutUnbound([-175.0, 175.0])
    )

    @property
    def l(self):
        """L channel."""

        return self._coords[0]

    @l.setter
    def l(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def u(self):
        """U channel."""

        return self._coords[1]

    @u.setter
    def u(self, value):
        """U axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def v(self):
        """V channel."""

        return self._coords[2]

    @v.setter
    def v(self, value):
        """V axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def _to_xyz(cls, parent, luv):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, luv_to_xyz(luv, cls.WHITE))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return xyz_to_luv(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz), cls.WHITE)



"""
Delta E ITP.

https://kb.portrait.com/help/ictcp-color-difference-metric
"""
from ..distance import DeltaE
import math
from ... import util


class DEITP(DeltaE):
    """Delta E ITP class."""

    @staticmethod
    def name():
        """Name of method."""

        return "itp"

    @staticmethod
    def distance(color, sample, scalar=720, **kwargs):
        """Delta E ITP color distance formula."""

        i1, t1, p1 = util.no_nan(color.convert('ictcp').coords())
        i2, t2, p2 = util.no_nan(sample.convert('ictcp').coords())

        # Equation (1)
        return scalar * math.sqrt((i1 - i2) ** 2 + 0.25 * (t1 - t2) ** 2 + (p1 - p2) ** 2)



"""Fit by compressing chroma in Lch."""
from ..gamut import Fit

EPSILON = 0.001


class LchChroma(Fit):
    """Lch chroma gamut mapping class."""

    @staticmethod
    def name():
        """Get plugin name."""

        return "lch-chroma"

    @staticmethod
    def fit(color):
        """
        Gamut mapping via chroma Lch.

        Algorithm originally came from https://colorjs.io/docs/gamut-mapping.html.
        Some things have been optimized and fixed though to better perform as intended.

        The idea is to hold hue and lightness constant and decrease chroma until
        color comes under gamut.

        We'll use a binary search and at after each stage, we will clip the color
        and compare the distance of the two colors (clipped and current color via binary search).
        If the distance is less than two, we can return the clipped color.

        ---
        Original Authors: Lea Verou, Chris Lilley
        License: MIT (As noted in https://github.com/LeaVerou/color.js/blob/master/package.json)
        """

        space = color.space()

        # If flooring chroma doesn't work, just clip the floored color
        # because there is no optimal compression.
        floor = color.clone().set('lch.chroma', 0)
        if not floor.in_gamut(tolerance=0):
            return floor.fit(method="clip").coords()

        # If we are already below the JND, just clip as we will gain no
        # noticeable difference moving forward.
        clipped = color.fit(method="clip")
        if color.delta_e(clipped, method="2000") < 2:
            return clipped.coords()

        # Convert to CIELCH and set our boundaries
        mapcolor = color.convert("lch")
        low = 0.0
        high = mapcolor.chroma

        # Adjust chroma (using binary search).
        # This helps preserve the other attributes of the color.
        # Each time we compare the compressed color to it's clipped form
        # to see how close we are. A delta less than 2 is our target.
        while (high - low) > EPSILON:
            delta = mapcolor.delta_e(
                mapcolor.fit(space, method="clip"),
                method="2000"
            )

            if (delta - 2) < EPSILON:
                low = mapcolor.chroma
            else:
                high = mapcolor.chroma

            mapcolor.chroma = (high + low) * 0.5

        # Update and clip off noise
        return color.update(mapcolor).fit(space, method="clip", in_place=True).coords()



"""Din99o Lch class."""
from ..spaces import RE_DEFAULT_MATCH
from .lch.base import LchBase
from .din99o import Din99o
from .. import util
import math
import re

ACHROMATIC_THRESHOLD = 0.0000000002


def lch_to_lab(lch):
    """Din99o Lch to lab."""

    l, c, h = lch
    h = util.no_nan(h)

    # If, for whatever reason (mainly direct user input),
    # if chroma is less than zero, clamp to zero.
    if c < 0.0:
        c = 0.0

    return [
        l,
        c * math.cos(math.radians(h)),
        c * math.sin(math.radians(h))
    ]


def lab_to_lch(lab):
    """Din99o Lab to Lch."""

    l, a, b = lab
    h = math.degrees(math.atan2(b, a))
    c = math.sqrt(a ** 2 + b ** 2)

    # Achromatic colors will often get extremely close, but not quite hit zero.
    # Essentially, we want to discard noise through rounding and such.
    if c <= ACHROMATIC_THRESHOLD:
        h = util.NaN

    return [l, c, util.constrain_hue(h)]


class Din99oLch(LchBase):
    """Lch D65 class."""

    SPACE = "din99o-lch"
    SERIALIZE = ("--din99o-lch",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def _to_din99o(cls, parent, lch):
        """To Lab."""

        return lch_to_lab(lch)

    @classmethod
    def _from_din99o(cls, parent, lab):
        """To Lab."""

        return lab_to_lch(lab)

    @classmethod
    def _to_xyz(cls, parent, lch):
        """To XYZ."""

        return Din99o._to_xyz(parent, cls._to_din99o(parent, lch))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return cls._from_din99o(parent, Din99o._from_xyz(parent, xyz))



"""
Oklab class.

https://bottosson.github.io/posts/oklab/
"""
from ..spaces import Space, RE_DEFAULT_MATCH, GamutUnbound, OptionalPercent, Labish
from .xyz import XYZ
from .. import util
import re

m1 = [
    [0.8189330101, 0.0329845436, 0.0482003018],
    [0.3618667424, 0.9293118715, 0.2643662691],
    [-0.1288597137, 0.0361456387, 0.633851707]
]

m2 = [
    [0.2104542553, 1.9779984951, 0.0259040371],
    [0.793617785, -2.428592205, 0.7827717662],
    [-0.0040720468, 0.4505937099, -0.808675766]
]

m1i = [
    [1.2270138511035211, -0.04058017842328059, -0.07638128450570689],
    [-0.5577999806518223, 1.11225686961683, -0.42148197841801266],
    [0.2812561489664678, -0.0716766786656012, 1.5861632204407947]
]

m2i = [
    [0.9999999984505199, 1.0000000088817607, 1.0000000546724108],
    [0.3963377921737679, -0.10556134232365635, -0.08948418209496575],
    [0.2158037580607588, -0.06385417477170591, -1.2914855378640917]
]


def xyz_d65_to_oklab(xyzd65):
    """XYZ D65 to Oklab."""

    return util.dot([util.cbrt(x) for x in util.dot(xyzd65, m1)], m2)


def oklab_to_xyz_d65(oklab):
    """From XYZ to LMS."""

    return util.dot([x ** 3 for x in util.dot(oklab, m2i)], m1i)


class Oklab(Labish, Space):
    """Oklab class."""

    SPACE = "oklab"
    SERIALIZE = ("--oklab",)
    CHANNEL_NAMES = ("l", "a", "b", "alpha")
    CHANNEL_ALIASES = {
        "lightness": "l"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    RANGE = (
        GamutUnbound([OptionalPercent(0), OptionalPercent(1)]),
        GamutUnbound([-0.5, 0.5]),
        GamutUnbound([-0.5, 0.5])
    )

    @property
    def l(self):
        """L channel."""

        return self._coords[0]

    @l.setter
    def l(self, value):
        """Get true luminance."""

        self._coords[0] = self._handle_input(value)

    @property
    def a(self):
        """A channel."""

        return self._coords[1]

    @a.setter
    def a(self, value):
        """A axis."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self):
        """B channel."""

        return self._coords[2]

    @b.setter
    def b(self, value):
        """B axis."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def _to_xyz(cls, parent, oklab):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, oklab_to_xyz_d65(oklab))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return xyz_d65_to_oklab(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz))



"""Test HSV library."""
import unittest
from . import util
from coloraide import Color, NaN


class TestHSVInputOutput(util.ColorAsserts, unittest.TestCase):
    """Test HSV."""

    def test_input_raw(self):
        """Test raw input."""

        self.assertColorEqual(Color("hsv", [20, 100, 75]), Color('color(--hsv 20 100% 75%)'))

    def test_color_class(self):
        """Test raw input."""

        self.assertColorEqual(Color(Color("hsv", [20, 100, 75])), Color('color(--hsv 20 100% 75%)'))

    def test_color(self):
        """Test color input/output format."""

        args = {"color": True}
        color = "color(--hsv 20 100% 75%)"

        self.assertEqual(Color(color).to_string(**args), 'color(--hsv 20 100% 75%)')

        color = "color(--hsv 20 100% 75% / 0.5)"
        self.assertEqual(Color(color).to_string(**args), 'color(--hsv 20 100% 75% / 0.5)')

        color = "color(--hsv 20 100% 75% / 50%)"
        self.assertEqual(Color(color).to_string(**args), 'color(--hsv 20 100% 75% / 0.5)')

    def test_no_alpha(self):
        """Test no alpha."""

        args = {"alpha": False}

        color = "color(--hsv 20 100% 75% 0.2)"
        hsv = Color(color)
        self.assertEqual("color(--hsv 20 100% 75%)", hsv.to_string(**args))

    def test_force_alpha(self):
        """Test force alpha."""

        args = {"alpha": True}

        color = "color(--hsv 20 100% 75% / 1)"
        hsv = Color(color)
        self.assertEqual("color(--hsv 20 100% 75% / 1)", hsv.to_string(**args))

    def test_precision(self):
        """Test precision."""

        color = 'color(--hsv 20.1234567 50.1234567% 50.1234567%)'
        self.assertEqual(Color(color).to_string(), 'color(--hsv 20.123 50.123% 50.123%)')
        self.assertEqual(Color(color).to_string(precision=3), 'color(--hsv 20.1 50.1% 50.1%)')
        self.assertEqual(Color(color).to_string(precision=0), 'color(--hsv 20 50% 50%)')
        self.assertEqual(
            Color(color).to_string(precision=-1),
            'color(--hsv 20.12345669999999842048055143095552921295166015625 50.12345669999999842048055143095552921295166015625% 50.12345669999999842048055143095552921295166015625%)'  # noqa:  E501
        )

    def test_fit(self):
        """Test fit."""

        self.assertEqual(
            Color('color(--hsv 20 150% 75%)').to_string(),
            'color(--hsv 40.492 100% 50.529%)'
        )

        self.assertEqual(
            Color('color(--hsv 20 150% 75%)').to_string(fit="clip"),
            'color(--hsv 20 100% 75%)'
        )

        self.assertEqual(
            Color('color(--hsv 20 150% 75%)').to_string(fit=False),
            'color(--hsv 20 150% 75%)'
        )


class TestHSVProperties(util.ColorAsserts, unittest.TestCase):
    """Test HSV."""

    def test_hue(self):
        """Test `hue`."""

        c = Color('color(--hsv 120 50% 50% / 1)')
        self.assertEqual(c.hue, 120)
        c.hue = 110
        self.assertEqual(c.hue, 110)

    def test_saturation(self):
        """Test `saturation`."""

        c = Color('color(--hsv 120 50% 50% / 1)')
        self.assertEqual(c.saturation, 50)
        c.saturation = 60
        self.assertEqual(c.saturation, 60)

    def test_value(self):
        """Test `value`."""

        c = Color('color(--hsv 120 50% 50% / 1)')
        self.assertEqual(c.value, 50)
        c.value = 40
        self.assertEqual(c.value, 40)

    def test_alpha(self):
        """Test `alpha`."""

        c = Color('color(--hsv 120 50% 50% / 1)')
        self.assertEqual(c.alpha, 1)
        c.alpha = 0.5
        self.assertEqual(c.alpha, 0.5)


class TestNull(util.ColorAsserts, unittest.TestCase):
    """Test Null cases."""

    def test_null_input(self):
        """Test null input."""

        c = Color('hsv', [NaN, 50, 75], 1)
        self.assertTrue(c.is_nan('hue'))

    def test_auto_null(self):
        """Test auto null."""

        c = Color('color(--hsv 120 0% 75% / 1)')
        self.assertTrue(c.is_nan('hue'))

    def test_to_hsl(self):
        """Test null from Lab conversion."""

        c1 = Color('color(--hsv 0 0% 50%)')
        c2 = c1.convert('hsl')
        self.assertColorEqual(c2, Color('hsl(0 0% 50%)'))
        self.assertTrue(c2.is_nan('hue'))

    def test_from_hsl(self):
        """Test null from Lab conversion."""

        c1 = Color('hsl(0 0% 50%)')
        c2 = c1.convert('hsv')
        self.assertColorEqual(c2, Color('color(--hsv 0 0% 50%)'))
        self.assertTrue(c2.is_nan('hue'))



"""HWB class."""
from ...spaces import Space, RE_DEFAULT_MATCH, Angle, Percent, GamutBound, Cylindrical
from ..srgb.base import SRGB
from ..hsv import HSV
from ... import util
import re


def hwb_to_hsv(hwb):
    """HWB to HSV."""

    h, w, b = hwb
    w /= 100.0
    b /= 100.0

    wb = w + b
    if (wb >= 1):
        gray = w / wb
        return [util.NaN, 0.0, gray * 100.0]

    v = 1 - b
    s = 0 if v == 0 else 1 - w / v
    return [h, s * 100, v * 100]


def hsv_to_hwb(hsv):
    """HSV to HWB."""

    h, s, v = hsv
    s /= 100
    v /= 100
    w = v * (1 - s)
    b = 1 - v
    if w + b >= 1:
        h = util.NaN
    return [h, w * 100, b * 100]


class HWB(Cylindrical, Space):
    """HWB class."""

    SPACE = "hwb"
    SERIALIZE = ("--hwb",)
    CHANNEL_NAMES = ("h", "w", "b", "alpha")
    CHANNEL_ALIASES = {
        "hue": "h",
        "whiteness": "w",
        "blackness": "b"
    }
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    GAMUT_CHECK = "srgb"
    WHITE = "D65"

    RANGE = (
        GamutBound([Angle(0.0), Angle(360.0)]),
        GamutBound([Percent(0.0), Percent(100.0)]),
        GamutBound([Percent(0.0), Percent(100.0)])
    )

    @property
    def h(self):
        """Hue channel."""

        return self._coords[0]

    @h.setter
    def h(self, value):
        """Shift the hue."""

        self._coords[0] = self._handle_input(value)

    @property
    def w(self):
        """Whiteness channel."""

        return self._coords[1]

    @w.setter
    def w(self, value):
        """Set whiteness channel."""

        self._coords[1] = self._handle_input(value)

    @property
    def b(self):
        """Blackness channel."""

        return self._coords[2]

    @b.setter
    def b(self, value):
        """Set blackness channel."""

        self._coords[2] = self._handle_input(value)

    @classmethod
    def null_adjust(cls, coords, alpha):
        """On color update."""

        if coords[1] + coords[2] >= 100:
            coords[0] = util.NaN
        return coords, alpha

    @classmethod
    def _to_xyz(cls, parent, hwb):
        """SRGB to XYZ."""

        return SRGB._to_xyz(parent, cls._to_srgb(parent, hwb))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """XYZ to SRGB."""

        return cls._from_srgb(parent, SRGB._from_xyz(parent, xyz))

    @classmethod
    def _to_srgb(cls, parent, hwb):
        """To sRGB."""

        return HSV._to_srgb(parent, cls._to_hsv(parent, hwb))

    @classmethod
    def _from_srgb(cls, parent, srgb):
        """From sRGB."""

        return cls._from_hsv(parent, HSV._from_srgb(parent, srgb))

    @classmethod
    def _to_hsl(cls, parent, hwb):
        """To HSL."""

        return HSV._to_hsl(parent, hwb_to_hsv(hwb))

    @classmethod
    def _from_hsl(cls, parent, hsl):
        """From HSL."""

        return hsv_to_hwb(HSV._from_hsl(parent, hsl))

    @classmethod
    def _to_hsv(cls, parent, hwb):
        """To HSV."""

        return hwb_to_hsv(hwb)

    @classmethod
    def _from_hsv(cls, parent, hsv):
        """From HSV."""

        return hsv_to_hwb(hsv)



"""Convert utilities."""
from ... import util
from . import cat


class Convert:
    """Conversion methods."""

    def chromatic_adaptation(self, w1, w2, xyz):
        """Apply chromatic adaption to XYZ coordinates."""

        method = self.CHROMATIC_ADAPTATION
        return cat.chromatic_adaptation(w1, w2, xyz, method=method)

    def convert(self, space, *, fit=False, in_place=False):
        """Convert to color space."""

        space = space.lower()

        if fit:
            method = None if not isinstance(fit, str) else fit
            if not self.in_gamut(space, tolerance=0.0):
                converted = self.convert(space, in_place=in_place)
                return converted.fit(space, method=method, in_place=True)

        convert_to = '_to_{}'.format(space)
        convert_from = '_from_{}'.format(self.space())

        obj = self.CS_MAP.get(space)
        if obj is None:
            raise ValueError("'{}' is not a valid color space".format(space))

        # See if there is a direct conversion route
        func = None
        if self.space() != space:
            # Don't send NaNs
            coords = util.no_nan(self.coords())
            if hasattr(self._space, convert_to):
                func = getattr(self._space, convert_to)
                coords = func(self, coords)
            elif hasattr(obj, convert_from):
                func = getattr(obj, convert_from)
                coords = func(self, coords)

            # See if there is an XYZ route
            if func is None and self.space() != space:
                func = getattr(self._space, '_to_xyz')
                coords = func(self, coords)

                if space != 'xyz':
                    func = getattr(obj, '_from_xyz')
                    coords = func(self, coords)
        else:
            # Nothing to convert, just pass values as is
            coords = self.coords()

        return self.mutate(space, coords, self.alpha) if in_place else self.new(space, coords, self.alpha)

    def mutate(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Mutate the current color to a new color."""

        c = self.new(color, data=data, alpha=alpha, filters=filters, **kwargs)
        self._attach(c._space)
        return self

    def update(self, color, data=None, alpha=util.DEF_ALPHA, *, filters=None, **kwargs):
        """Update the existing color space with the provided color."""

        c = self.new(color, data=data, alpha=alpha, filters=filters, **kwargs)
        if c.space() != self.space():
            c.convert(self.space(), in_place=True)
        self._attach(c._space)
        return self



"""Color contrast."""


class Contrast:
    """Contrast."""

    def luminance(self):
        """Get color's luminance."""

        return self.convert("xyz-d65").y

    def contrast(self, color):
        """Compare the contrast ratio of this color and the provided color."""

        color = self._handle_color_input(color)
        lum1 = self.luminance()
        lum2 = color.luminance()
        return (lum1 + 0.05) / (lum2 + 0.05) if (lum1 > lum2) else (lum2 + 0.05) / (lum1 + 0.05)



"""Distance and Delta E."""
from . import distance_euclidean
from abc import ABCMeta, abstractmethod


class DeltaE(ABCMeta):
    """Delta E plugin class."""

    @staticmethod
    @abstractmethod
    def name():
        """Get name of method."""

    @staticmethod
    @abstractmethod
    def distance(color, sample, **kwargs):
        """Get distance between color and sample."""


class Distance:
    """Distance."""

    def delta_e(self, color, *, method=None, **kwargs):
        """Delta E distance."""

        color = self._handle_color_input(color)
        if method is None:
            method = self.DELTA_E

        algorithm = method.lower()

        try:
            return self.DE_MAP[algorithm](self, color, **kwargs)
        except KeyError:
            raise ValueError("'{}' is not currently a supported distancing algorithm.".format(algorithm))

    def distance(self, color, *, space="lab"):
        """Delta."""

        color = self._handle_color_input(color)
        return distance_euclidean.distance(self, color, space=space)



"""SRGB Linear color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb.base import SRGB, lin_srgb_to_xyz, xyz_to_lin_srgb, lin_srgb, gam_srgb
from .xyz import XYZ
import re


class SRGBLinear(SRGB):
    """SRGB linear."""

    SPACE = "srgb-linear"
    SERIALIZE = ("--srgb-linear",)
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space='|'.join(SERIALIZE), channels=3))
    WHITE = "D65"

    @classmethod
    def _to_srgb(cls, parent, rgb):
        """Linear sRGB to sRGB."""

        return gam_srgb(rgb)

    @classmethod
    def _from_srgb(cls, parent, rgb):
        """sRGB to linear sRGB."""

        return lin_srgb(rgb)

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """SRGB Linear to XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_srgb_to_xyz(rgb))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """XYZ to SRGB Linear."""

        return xyz_to_lin_srgb(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz))



"""Pro Photo RGB color class."""
from ..spaces import RE_DEFAULT_MATCH
from .srgb.base import SRGB
from .xyz import XYZ
from .. import util
import re

ET = 1 / 512
ET2 = 16 / 512

RGB_TO_XYZ = [
    [7.9776048967230273e-01, 1.3518583717574034e-01, 3.1349349581524806e-02],
    [2.8807112822929343e-01, 7.1184321781010151e-01, 8.5653960605259035e-05],
    [0.0000000000000000e+00, 0.0000000000000000e+00, 8.2510460251046025e-01]
]

XYZ_TO_RGB = [
    [1.345798973102828, -0.2555801000799754, -0.05110628506753401],
    [-0.5446224939028346, 1.508232741313278, 0.020536032391479723],
    [0.0, 0.0, 1.2119675456389452]
]


def lin_prophoto_to_xyz(rgb):
    """
    Convert an array of linear-light prophoto-rgb values to CIE XYZ using  D50.D50.

    (so no chromatic adaptation needed afterwards)
    http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html
    """

    return util.dot(RGB_TO_XYZ, rgb)


def xyz_to_lin_prophoto(xyz):
    """Convert XYZ to linear-light prophoto-rgb."""

    return util.dot(XYZ_TO_RGB, xyz)


def lin_prophoto(rgb):
    """
    Convert an array of prophoto-rgb values in the range 0.0 - 1.0 to linear light (un-corrected) form.

    Transfer curve is gamma 1.8 with a small linear portion.

    https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        if abs(i) < ET2:
            result.append(i / 16.0)
        else:
            result.append(util.npow(i, 1.8))
    return result


def gam_prophoto(rgb):
    """
    Convert an array of linear-light prophoto-rgb  in the range 0.0-1.0 to gamma corrected form.

    Transfer curve is gamma 1.8 with a small linear portion.

    https://en.wikipedia.org/wiki/ProPhoto_RGB_color_space
    """

    result = []
    for i in rgb:
        # Mirror linear nature of algorithm on the negative axis
        if abs(i) < ET:
            result.append(16.0 * i)
        else:
            result.append(util.nth_root(i, 1.8))
    return result


class ProPhotoRGB(SRGB):
    """Pro Photo RGB class."""

    SPACE = "prophoto-rgb"
    DEFAULT_MATCH = re.compile(RE_DEFAULT_MATCH.format(color_space=SPACE, channels=3))
    WHITE = "D50"

    @classmethod
    def _to_xyz(cls, parent, rgb):
        """To XYZ."""

        return parent.chromatic_adaptation(cls.WHITE, XYZ.WHITE, lin_prophoto_to_xyz(lin_prophoto(rgb)))

    @classmethod
    def _from_xyz(cls, parent, xyz):
        """From XYZ."""

        return gam_prophoto(xyz_to_lin_prophoto(parent.chromatic_adaptation(XYZ.WHITE, cls.WHITE, xyz)))
